##################################
#    CMFSubversionBrowser
#    Copyright (C) 2002-2004 kapil thangavelu <k_vertigo@objectrealms.net>
#
#    This product is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This product is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this product; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#################################

"""
Purpose: Initialize and Register the Browser with Zope/CMF
Author: kapil thangavelu <pluto@objectreamls.net>
$Id: SubversionBrowser.py,v 1.3 2003/02/28 05:14:58 k_vertigo Exp $

Details

    - Offer a browser into an svn repository via zope

    - Explore svn apis

this module can be imported outside of zope provided the support modules
are on the PYTHONPATH.
"""

import atexit, os, sys, difflib, time

# our utilities
from utils import svn_path_join, charbuffer, wrap

# svn bindings
from svn import fs, repos, util

# zope infrastructure
from Acquisition import Implicit, aq_base, aq_parent
from AccessControl import ClassSecurityInfo, Permissions
from DateTime import DateTime
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from Shared.DC.ZRDB.TM import TM
from zLOG import LOG
from Persistence import Persistent


#################################
# Globals
DEBUG = 0
CHUNK_SIZE = 16384
_marker = object()
#################################
# log class
class _log:

    LOG_NAME = '[SubversionBrowser]'
    LOG_LEVEL = 100

    def debug(self, msg, extra='', **kwargs):
        if not DEBUG:
            return
        #print msg
        LOG(self.LOG_NAME,
            self.LOG_LEVEL,
            msg,
            extra)

log = _log()

#################################
# APR Jaunx
#
# XXX we should check if we've already been registered

APR_NEEDS_INIT = 1
APR_NEEDS_EXIT = 1

if APR_NEEDS_INIT:
    log.debug('Initializing APR')
    util.apr_initialize()

if APR_NEEDS_EXIT:
    def apr_exit():
        util.apr_terminate()
    atexit.register(apr_exit)

#################################
# SVN Custodial Engineer 

class svn_sweeper:
    """
    cleans up browser svn resources.
    """
    def __init__(self, obj):
        self.obj = obj

    def sweep(self):
        if self.obj._v_initialized == None:
            return
        fs.close_root(self.obj._v_fsroot)             # close revision
        # XXX as of .18 didn't get exported for me.. strange
        #repos.svn_repos_close(self.obj._v_repository) # close repository
        util.svn_pool_clear(self.obj._v_pool)         # clean pool
        util.svn_pool_destroy(self.obj._v_pool)       # destroy pool
        self.obj._v_initialized = None
        
#################################
# svn browser

class InvalidRepositoryPath(Exception): pass

class SubversionBrowser(SimpleItem):

    meta_type = 'Subversion Browser'

    security = ClassSecurityInfo()

    # volatile variables for holding svn resources
    _v_pool = None
    _v_repository = None
    _v_fsptr = None
    _v_fsroot = None
    _v_revision = None
    _v_initialized = None

    def __init__(self, id, repository_path='', svnpath=''):
        self.id = id
        self.repository_path = repository_path
        self.svnpath = svnpath  # the path we should start browsing at
        
    security.declarePrivate('initialize')
    def initialize(self, revision=None):
        """
        initialize the browser, with a new pool, revison root, and repository
        pointer. optional revision arg can specify the revision root
        """
        if self._v_initialized and ((revision is None) or revision==self._v_revision):
            return 

        # clean up so we can reinitialize with a different root
        if self._v_initialized:
            svn_sweeper(self).sweep()
            
        if revision is not None:
            assert int(revision), " rev's must be positive integers"

        log.debug('initializing')
        
        # XXXXX we should be checking for used pools and cleaning them.
        
        self._v_pool       = util.svn_pool_create(None)
        self._v_repository = repos.svn_repos_open(self.repository_path, self._v_pool)
        self._v_fsptr      = repos.svn_repos_fs(self._v_repository)
        
        if revision is None:
            self._v_revision = fs.youngest_rev(self._v_fsptr, self._v_pool)
        else:
            self._v_revision = revision
            
        self._v_fsroot = fs.revision_root(self._v_fsptr,
                                          self._v_revision,
                                          self._v_pool)
        self._v_initialized = 1

    security.declarePrivate('setRevision')
    def setRevision(self, revision=None):
        fs.close_root(self._v_fsroot)             # close revision
        self._v_revision = int(revision)
        self._v_fsroot = fs.revision_root(self._v_fsptr,
                                          self._v_revision,
                                          self._v_pool)
                                          
        
    security.declarePrivate('clear')
    def clear(self):
        svn_sweeper(self).sweep()
        
    security.declarePrivate('getSVNObject')
    def getSVNObject(self, path):
        """
        retrieve a subversion object from an svn path. does
        not wrap the object.
        """
        log.debug('retrieving %s'%str(path))
        
        if not self._v_initialized:
            self.initialize()
            
        if fs.is_dir(self._v_fsroot, path, self._v_pool):
            return SVNDirectoryFactory(path)
        elif fs.is_file(self._v_fsroot, path, self._v_pool):
            return SVNFileFactory(path)
        else:
            return None

    security.declarePrivate('getSVNBrowswer')
    def getSVNBrowser(self):
        return self

    #################################
    # public api methods

    security.declareProtected(Permissions.view, 'getRootSVNObject')
    def getSVNRootObject(self):
        return self.getSVNObject(self.svnpath).__of__(self)


    def editBrowser(self, repository_path, svnpath):
        if not os.path.exists(repository_path):
            raise InvalidRepositoryPath(repository_path)
        
        self.repository_path = repository_path
        self.svnpath = svnpath
        
        # clean up existing conns and memory pools
        svn_sweeper(self).sweep()

        # try and open the repository just to be sure
        try:
            self.initialize()
        finally:
            svn_sweeper(self).sweep()
        
    #################################

    def __getstate__(self):
        # do not go dirty into that good night
        svn_sweeper(self).sweep()
        return Persistent.__getstate__(self)

InitializeClass(SubversionBrowser)

#################################
# svn object model        

class SubversionObject(Implicit):

    security = ClassSecurityInfo()
    
    def __init__(self, id=None, svnpath=None):
        self.id = id
        self.svnpath = svnpath

    security.declarePublic('getId')
    def getId(self):
        return self.id

    security.declarePrivate('getNodeId')
    def getNodeId(self):
        """
        Return the Subversion Node Id
        """
        browser = self.getSVNBrowser()
        browser.initialize()
        uid = fs.node_id(browser._v_fsroot, self.svnpath, browser._v_pool)
        return '%s'%fs.unparse_id(uid, browser._v_pool)

    security.declareProtected(Permissions.view, 'getLog')
    def getLog(self, revision=None):
        """ 
        Return A Log Entry Object for A Single Revision. If Revision
        is not specified then the revision is the latest available.
        """
        browser = self.getSVNBrowser()
        browser.initialize()
        
        if revision is None:
            revision = fs.node_created_rev(browser._v_fsroot,
                                           self.svnpath,
                                           browser._v_pool)
        
        return SVNLogFactory(browser._v_fsptr, revision, browser._v_pool)

    security.declareProtected(Permissions.view, 'getLogEntries')
    def getLogEntries(self):
        """
        get all log entries associated with this node
        """
        
        browser = self.getSVNBrowser()
        browser.initialize()
        collection = SVNLogCollection()
        
        repos.svn_repos_get_logs(
            browser._v_repository,  # repository
            [self.svnpath],         # paths
            browser._v_revision,    # start revision
            0,                      # end revision (for sort)
            0,                      # discover changed paths 
            0,                      # strict node history
            collection.addLog,      # reciever
            browser._v_pool)

        return collection.getLogs()

    def getParent(self):
        """
        get the parent object of this node
        XXX check security of browser's root svnpath as limiting factor
        """
        browser = self.getSVNBrowser()
        segments = self.svnpath.split('/')
        segments.pop(-1)
        return browser.getSVNObject(svn_path_join(*segments)).__of__(browser)

    security.declarePublic('getProperty')
    def getProperty(self, property_name): 
        """
        retrieve a given property name
        """
        browser = self.getSVNBrowser()
        browser.initialize()
        
        return fs.node_prop(browser._v_fsroot, # root
                            self.svnpath,     # path 
                            property_name,    # prop name
                            browser._v_pool   # pool
                            )

    security.declarePublic('getProperties')
    def getProperties(self):
        """
        return all the node property names
        """
        browser = self.getSVNBrowser()
        browser.initialize()
        
        return fs.node_proplist(browser._v_fsroot,
                            self.svnpath,
                            browser._v_pool)


    def getRevisionPathMap(self, cross_copies=0):
        """
        for every revision this node was modified return a dictionary
        mapping each revision to the path the node was at.
        """
        browser = self.getSVNBrowser()
        browser.initialize()
        pool = util.svn_pool_create(browser._v_pool)
        rev_map = {}
        history = fs.node_history( browser._v_fsroot, self.svnpath, pool)

        try:
            while history:
                history = fs.history_prev( history, cross_copies, pool )
                if history is None: break
                path, rev = fs.history_location( history, pool )
                rev_map[ rev ] = path
        finally:
            util.svn_pool_destroy( pool )
            
        return rev_map
        
    def getRevisionCreated(self):
        """
        return the revision that this node was created.
        """
        browser = self.getSVNBrowser()
        browser.initialize()

        pool = util.svn_pool_create(browser._v_pool)
        try:
            revision = fs.node_created_rev(browser._v_fsroot,
                                           self.svnpath,
                                           pool)
            
        finally:
            util.svn_pool_destroy(pool)

        return revision
    
    security.declarePublic('isBinary')
    def isBinary(self):

        mime_type = self.getProperty(util.SVN_PROP_MIME_TYPE)
        if mime_type is None:
            return 0        
        return util.svn_mime_type_is_binary(mime_type)
        
    security.declarePublic('absolute_url')
    def absolute_url(self, relative=0):
        
        browser = self.getSVNBrowser()
        url = browser.absolute_url(relative=relative)
        relative_path = self.svnpath[len(browser.svnpath)+1:]
        
        return "%s/%s"%(url, relative_path)
        

InitializeClass(SubversionObject)

class SubversionDirectory(SubversionObject):
    """
    Subversion Directory Node
    """
    
    meta_type = 'Subversion Directory'

    security = ClassSecurityInfo()

    isSVNDirectory = 1

    def __getitem__(self, key):

        browser = self.getSVNBrowser()
        path = svn_path_join(self.svnpath, key)

        try:
            if fs.is_dir(browser._v_fsroot, path, browser._v_pool):
                return SVNDirectoryFactory(path).__of__(self)
            if fs.is_file(browser._v_fsroot, path, browser._v_pool):
                return SVNFileFactory(path).__of__(self)
        except:
            return getattr(self, key)

    security.declareProtected(Permissions.view, 'getDirectories')
    def getDirectories(self):
        
        browser = self.getSVNBrowser()

        entries = fs.dir_entries(browser._v_fsroot,
                                 self.svnpath,
                                 browser._v_pool)

        names = entries.keys()
        names.sort()
        paths = map(lambda n, _p=self.svnpath: svn_path_join(_p, n), names)
        
        pool = util.svn_pool_create(browser._v_pool)
        
        try:
            dirs = [d for d in paths if fs.is_dir(browser._v_fsroot, d, pool)]
        finally:
            util.svn_pool_clear(pool)
            util.svn_pool_destroy(pool)
        
        return map( wrap(self),
                    map(SVNDirectoryFactory, dirs) ) 

    security.declareProtected(Permissions.view, 'getFiles')
    def getFiles(self):

        browser = self.getSVNBrowser()

        entries = fs.dir_entries(browser._v_fsroot,
                                 self.svnpath,
                                 browser._v_pool)

        names = entries.keys()
        names.sort()
        paths = map(lambda n, _p=self.svnpath: svn_path_join(_p, n), names)

        pool = util.svn_pool_create(browser._v_pool)
        
        try:
            files = [d for d in paths if fs.is_file(browser._v_fsroot, d, pool)]
        finally:
            util.svn_pool_clear(pool)
            util.svn_pool_destroy(pool)
            
        return map( wrap(self),
                    map(SVNFileFactory, files) )

InitializeClass(SubversionDirectory)

class SubversionFile(SubversionObject):
    """
    Subversion File Node
    """
    
    meta_type = 'Subversion File'
    
    security = ClassSecurityInfo()

    isSVNFile = 1

    security.declareProtected(Permissions.view, 'getContents')
    def getContents(self, revision=None, writer=None):

        rval = not writer
        browser = self.getSVNBrowser()

        if revision is not None and isinstance(revision, int):
            browser.initialize(revision)
        
        stream = fs.file_contents(browser._v_fsroot,
                                  self.svnpath,
                                  browser._v_pool)

        if writer is None:
            writer = charbuffer()

        while 1:
            data = util.svn_stream_read(stream, CHUNK_SIZE)
            if not data:
                break
            writer.write(data)

        if revision is not None:
            browser.clear()
            

        if rval:
            return writer.getvalue()

    security.declarePublic('getMimeType')
    def getMimeType(self):

        mime_type = self.getProperty(util.SVN_PROP_MIME_TYPE)
        if mime_type is None:
            return 'text/plain'        
        return mime_type

    security.declareProtected(Permissions.view, 'getAnnotatedLines')
    def getAnnotatedLines(self, revision_set=(), include_copies=0):
        stime = time.time()
        browser = self.getSVNBrowser()
        annotations = []
        contents = ''

        revision_path_map = self.getRevisionPathMap( include_copies )

        if not revision_set:
            revision_set = revision_path_map.keys()

        # set revision order low high
        revision_set.sort()
        # flush mem.
        browser.initialize(revision_set[-1])

        if not len(revision_set) > 1 or self.isBinary():
            return ()

        pool = util.svn_pool_create(browser._v_pool)
        
        try:
            for r in revision_set:
                revision_root = fs.revision_root(browser._v_fsptr, r, pool)
                
                rev_author = fs.revision_prop(browser._v_fsptr,
                                              r,
                                              util.SVN_PROP_REVISION_AUTHOR,
                                              pool)
                
                stream = fs.file_contents(revision_root,
                                          revision_path_map[r],
                                          pool)
                buffer = charbuffer()

                while 1:
                    data = util.svn_stream_read(stream, CHUNK_SIZE)
                    if not data:
                        break
                    buffer.write(data)

                previous_contents = contents
                contents = buffer.getvalue()
                delta = difflib.ndiff(
                    previous_contents.splitlines(1),
                    contents.splitlines(1)
                    )

                # annotate algorithm
                li = 0
                for d in delta:
                    flag = d[0]
                    if flag == ' ':
                        li += 1
                        continue
                    elif flag == '?':
                        continue
                    elif flag == '+':
                        annotations.insert(li, (r, d[2:], rev_author) )
                        li += 1
                    elif flag == '-':
                        del annotations[li]

                fs.close_root(revision_root)
        finally:
            util.svn_pool_clear(pool)
            util.svn_pool_destroy(pool)

        # reset the browser
        svn_sweeper(browser).sweep()
        return [annotations[i] for i in xrange(len(annotations))]
            

InitializeClass(SubversionFile)        
        
class SubversionLogEntry:

    __allow_access_to_unprotected_subobjects__ = 1
    
    def __init__(self, **args):
        self.__dict__.update(args)

    def __repr__(self):
        return "<Log: %s>"%' '.join( self.__dict__.keys() )

#################################
# factories        

def SVNLogFactory(fsptr, revision, pool):
    
    author  = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_AUTHOR, pool)
    message = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_LOG, pool)
    date    = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_DATE, pool)
    
    date = util.svn_time_from_cstring(date, pool)

    return SubversionLogEntry(author = author,
                              message = message,
                              date = zope_time(date),
                              revision = revision
                              )

class SVNLogCollection:

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self):
        self.logs = []

    def addLog(self, paths, revision, author, date, message, pool):
        
        """
        Used for log gather callbacks
        
        notes in svn_repos.h, svn_types.h, svn_clients.h 
        
        typedef svn_error_t *(*svn_log_message_receiver_t)
        (void *baton,
        apr_hash_t *changed_paths,
        svn_revnum_t revision,
        const char *author,
        const char *date,  /* use svn_time_from_string() if need apr_time_t */
        const char *message,
        apr_pool_t *pool);
        """
        
        date = util.svn_time_from_cstring(date, pool)
    
        self.logs.append( SubversionLogEntry(author = author,
                                             message = message,
                                             date = zope_time(date),
                                             revision = revision,
                                             paths = paths
                                             )
                          )

    def __getitem__(self, i):
        return self.logs[i]
    

    def getLogs(self):
        return self.logs

# migrate to SVNUtil or id func on context
def SVNResourceFactory(svnpath, klass):

    idx=svnpath.rfind('/')
    if not idx:
        id = svnpath.strip()
    else:
        id = svnpath[idx+1:]

    return klass(id, svnpath)

def SVNDirectoryFactory(svnpath, klass=SubversionDirectory):
    return SVNResourceFactory(svnpath, klass)

def SVNFileFactory(svnpath, klass=SubversionFile):
    return SVNResourceFactory(svnpath, klass)


def zope_time(aprtime):
    return DateTime(aprtime/1000000)

if __name__ == '__main__':
    log.debug('svn browser starting')
    #o = SubversionBrowser('svn', '/opt/svn/repo/CMFDeployment')
    #o.initialize()
    #print o.getDirectories()
    
